package javax.swing.text;

import java.lang.ref.SoftReference;
import java.awt.Rectangle;
import java.awt.Container;
import java.awt.Shape;
import javax.swing.event.DocumentEvent;
import java.awt.Graphics;
import java.awt.Color;
import java.awt.FontMetrics;

public class WrappedPlainView extends BoxView implements TabExpander
{
    FontMetrics metrics;
    Segment lineBuffer;
    boolean widthChanging;
    int tabBase;
    int tabSize;
    boolean wordWrap;
    int sel0;
    int sel1;
    Color unselected;
    Color selected;
    
    public WrappedPlainView(Element element) {
        this(element, false);
    }
    
    public WrappedPlainView(Element element, boolean wordWrap) {
        super(element, 1);
        this.wordWrap = wordWrap;
    }
    
    protected int getTabSize() {
        Integer n = (Integer)this.getDocument().getProperty("tabSize");
        return (n != null) ? ((int)n) : 8;
    }
    
    protected void drawLine(int n, int n2, Graphics graphics, int drawText, int n3) {
        Element element = this.getElement();
        Element element2 = element.getElement(element.getElementIndex(n));
        try {
            if (element2.isLeaf()) {
                this.drawText(element2, n, n2, graphics, drawText, n3);
            }
            else {
                for (int i = element2.getElementIndex(n); i <= element2.getElementIndex(n2); ++i) {
                    Element element3 = element2.getElement(i);
                    drawText = this.drawText(element3, Math.max(element3.getStartOffset(), n), Math.min(element3.getEndOffset(), n2), graphics, drawText, n3);
                }
            }
        }
        catch (BadLocationException ex) {
            throw new StateInvariantError("Can't render: " + n + "," + n2);
        }
    }
    
    private int drawText(Element element, int n, int min, Graphics graphics, int n2, int n3) throws BadLocationException {
        min = Math.min(this.getDocument().getLength(), min);
        AttributeSet attributes = element.getAttributes();
        if (Utilities.isComposedTextAttributeDefined(attributes)) {
            graphics.setColor(this.unselected);
            n2 = Utilities.drawComposedText(this, attributes, graphics, n2, n3, n - element.getStartOffset(), min - element.getStartOffset());
        }
        else if (this.sel0 == this.sel1 || this.selected == this.unselected) {
            n2 = this.drawUnselectedText(graphics, n2, n3, n, min);
        }
        else if (n >= this.sel0 && n <= this.sel1 && min >= this.sel0 && min <= this.sel1) {
            n2 = this.drawSelectedText(graphics, n2, n3, n, min);
        }
        else if (this.sel0 >= n && this.sel0 <= min) {
            if (this.sel1 >= n && this.sel1 <= min) {
                n2 = this.drawUnselectedText(graphics, n2, n3, n, this.sel0);
                n2 = this.drawSelectedText(graphics, n2, n3, this.sel0, this.sel1);
                n2 = this.drawUnselectedText(graphics, n2, n3, this.sel1, min);
            }
            else {
                n2 = this.drawUnselectedText(graphics, n2, n3, n, this.sel0);
                n2 = this.drawSelectedText(graphics, n2, n3, this.sel0, min);
            }
        }
        else if (this.sel1 >= n && this.sel1 <= min) {
            n2 = this.drawSelectedText(graphics, n2, n3, n, this.sel1);
            n2 = this.drawUnselectedText(graphics, n2, n3, this.sel1, min);
        }
        else {
            n2 = this.drawUnselectedText(graphics, n2, n3, n, min);
        }
        return n2;
    }
    
    protected int drawUnselectedText(Graphics graphics, int n, int n2, int n3, int n4) throws BadLocationException {
        graphics.setColor(this.unselected);
        Document document = this.getDocument();
        Segment sharedSegment = SegmentCache.getSharedSegment();
        document.getText(n3, n4 - n3, sharedSegment);
        int drawTabbedText = Utilities.drawTabbedText(this, sharedSegment, n, n2, graphics, this, n3);
        SegmentCache.releaseSharedSegment(sharedSegment);
        return drawTabbedText;
    }
    
    protected int drawSelectedText(Graphics graphics, int n, int n2, int n3, int n4) throws BadLocationException {
        graphics.setColor(this.selected);
        Document document = this.getDocument();
        Segment sharedSegment = SegmentCache.getSharedSegment();
        document.getText(n3, n4 - n3, sharedSegment);
        int drawTabbedText = Utilities.drawTabbedText(this, sharedSegment, n, n2, graphics, this, n3);
        SegmentCache.releaseSharedSegment(sharedSegment);
        return drawTabbedText;
    }
    
    protected final Segment getLineBuffer() {
        if (this.lineBuffer == null) {
            this.lineBuffer = new Segment();
        }
        return this.lineBuffer;
    }
    
    protected int calculateBreakPosition(int n, int n2) {
        Segment sharedSegment = SegmentCache.getSharedSegment();
        this.loadText(sharedSegment, n, n2);
        int width = this.getWidth();
        int n3;
        if (this.wordWrap) {
            n3 = n + Utilities.getBreakLocation(sharedSegment, this.metrics, this.tabBase, this.tabBase + width, this, n);
        }
        else {
            n3 = n + Utilities.getTabbedTextOffset(sharedSegment, this.metrics, this.tabBase, this.tabBase + width, this, n, false);
        }
        SegmentCache.releaseSharedSegment(sharedSegment);
        return n3;
    }
    
    @Override
    protected void loadChildren(ViewFactory viewFactory) {
        Element element = this.getElement();
        int elementCount = element.getElementCount();
        if (elementCount > 0) {
            View[] array = new View[elementCount];
            for (int i = 0; i < elementCount; ++i) {
                array[i] = new WrappedPlainView.WrappedLine(element.getElement(i));
            }
            this.replace(0, 0, array);
        }
    }
    
    void updateChildren(DocumentEvent documentEvent, Shape shape) {
        DocumentEvent.ElementChange change = documentEvent.getChange(this.getElement());
        if (change != null) {
            Element[] childrenRemoved = change.getChildrenRemoved();
            Element[] childrenAdded = change.getChildrenAdded();
            View[] array = new View[childrenAdded.length];
            for (int i = 0; i < childrenAdded.length; ++i) {
                array[i] = new WrappedPlainView.WrappedLine(childrenAdded[i]);
            }
            this.replace(change.getIndex(), childrenRemoved.length, array);
            if (shape != null) {
                this.preferenceChanged(null, true, true);
                this.getContainer().repaint();
            }
        }
        this.updateMetrics();
    }
    
    final void loadText(Segment segment, int n, int n2) {
        try {
            this.getDocument().getText(n, n2 - n, segment);
        }
        catch (BadLocationException ex) {
            throw new StateInvariantError("Can't get line text");
        }
    }
    
    final void updateMetrics() {
        Container container = this.getContainer();
        this.metrics = container.getFontMetrics(container.getFont());
        this.tabSize = this.getTabSize() * this.metrics.charWidth('m');
    }
    
    @Override
    public float nextTabStop(float n, int n2) {
        if (this.tabSize == 0) {
            return n;
        }
        return (float)(this.tabBase + (((int)n - this.tabBase) / this.tabSize + 1) * this.tabSize);
    }
    
    @Override
    public void paint(Graphics graphics, Shape shape) {
        this.tabBase = ((Rectangle)shape).x;
        JTextComponent textComponent = (JTextComponent)this.getContainer();
        this.sel0 = textComponent.getSelectionStart();
        this.sel1 = textComponent.getSelectionEnd();
        this.unselected = (textComponent.isEnabled() ? textComponent.getForeground() : textComponent.getDisabledTextColor());
        this.selected = ((textComponent.getCaret().isSelectionVisible() && textComponent.getHighlighter() != null) ? textComponent.getSelectedTextColor() : this.unselected);
        graphics.setFont(textComponent.getFont());
        super.paint(graphics, shape);
    }
    
    @Override
    public void setSize(float n, float n2) {
        this.updateMetrics();
        if ((int)n != this.getWidth()) {
            this.preferenceChanged(null, true, true);
            this.widthChanging = true;
        }
        super.setSize(n, n2);
        this.widthChanging = false;
    }
    
    @Override
    public float getPreferredSpan(int n) {
        this.updateMetrics();
        return super.getPreferredSpan(n);
    }
    
    @Override
    public float getMinimumSpan(int n) {
        this.updateMetrics();
        return super.getMinimumSpan(n);
    }
    
    @Override
    public float getMaximumSpan(int n) {
        this.updateMetrics();
        return super.getMaximumSpan(n);
    }
    
    @Override
    public void insertUpdate(DocumentEvent documentEvent, Shape shape, ViewFactory viewFactory) {
        this.updateChildren(documentEvent, shape);
        Rectangle rectangle = (shape != null && this.isAllocationValid()) ? this.getInsideAllocation(shape) : null;
        View viewAtPosition = this.getViewAtPosition(documentEvent.getOffset(), rectangle);
        if (viewAtPosition != null) {
            viewAtPosition.insertUpdate(documentEvent, rectangle, viewFactory);
        }
    }
    
    @Override
    public void removeUpdate(DocumentEvent documentEvent, Shape shape, ViewFactory viewFactory) {
        this.updateChildren(documentEvent, shape);
        Rectangle rectangle = (shape != null && this.isAllocationValid()) ? this.getInsideAllocation(shape) : null;
        View viewAtPosition = this.getViewAtPosition(documentEvent.getOffset(), rectangle);
        if (viewAtPosition != null) {
            viewAtPosition.removeUpdate(documentEvent, rectangle, viewFactory);
        }
    }
    
    @Override
    public void changedUpdate(DocumentEvent documentEvent, Shape shape, ViewFactory viewFactory) {
        this.updateChildren(documentEvent, shape);
    }
    
    class WrappedLine extends View
    {
        int lineCount;
        SoftReference<int[]> lineCache;
        
        WrappedLine(Element element) {
            super(element);
            this.lineCache = null;
            this.lineCount = -1;
        }
        
        @Override
        public float getPreferredSpan(int i) {
            switch (i) {
                case 0: {
                    float n = (float)WrappedPlainView.this.getWidth();
                    if (n == 2.1474836E9f) {
                        return 100.0f;
                    }
                    return n;
                }
                case 1: {
                    if (this.lineCount < 0 || WrappedPlainView.this.widthChanging) {
                        this.breakLines(this.getStartOffset());
                    }
                    return (float)(this.lineCount * WrappedPlainView.this.metrics.getHeight());
                }
                default: {
                    throw new IllegalArgumentException("Invalid axis: " + i);
                }
            }
        }
        
        @Override
        public void paint(Graphics graphics, Shape shape) {
            Rectangle rectangle = (Rectangle)shape;
            int n = rectangle.y + WrappedPlainView.this.metrics.getAscent();
            int x = rectangle.x;
            JTextComponent textComponent = (JTextComponent)this.getContainer();
            Highlighter highlighter = textComponent.getHighlighter();
            LayeredHighlighter layeredHighlighter = (highlighter instanceof LayeredHighlighter) ? ((LayeredHighlighter)highlighter) : null;
            int startOffset = this.getStartOffset();
            int endOffset = this.getEndOffset();
            int[] lineEnds = this.getLineEnds();
            int n2;
            for (int i = 0; i < this.lineCount; startOffset = n2, n += WrappedPlainView.this.metrics.getHeight(), ++i) {
                n2 = ((lineEnds == null) ? endOffset : (startOffset + lineEnds[i]));
                if (layeredHighlighter != null) {
                    n2 = ((n2 == endOffset) ? (n2 - 1) : n2);
                    layeredHighlighter.paintLayeredHighlights(graphics, startOffset, n2, shape, textComponent, this);
                }
                WrappedPlainView.this.drawLine(startOffset, n2, graphics, x, n);
            }
        }
        
        @Override
        public Shape modelToView(int n, Shape shape, Position.Bias bias) throws BadLocationException {
            Rectangle bounds = shape.getBounds();
            bounds.height = WrappedPlainView.this.metrics.getHeight();
            bounds.width = 1;
            int startOffset = this.getStartOffset();
            if (n < startOffset || n > this.getEndOffset()) {
                throw new BadLocationException("Position out of range", n);
            }
            int n2 = (bias == Position.Bias.Forward) ? n : Math.max(startOffset, n - 1);
            int[] lineEnds = this.getLineEnds();
            if (lineEnds != null) {
                int line = this.findLine(n2 - startOffset);
                if (line > 0) {
                    startOffset += lineEnds[line - 1];
                }
                bounds.y += bounds.height * line;
            }
            if (n > startOffset) {
                Segment sharedSegment = SegmentCache.getSharedSegment();
                WrappedPlainView.this.loadText(sharedSegment, startOffset, n);
                bounds.x += Utilities.getTabbedTextWidth(sharedSegment, WrappedPlainView.this.metrics, bounds.x, WrappedPlainView.this, startOffset);
                SegmentCache.releaseSharedSegment(sharedSegment);
                return bounds;
            }
            return bounds;
        }
        
        @Override
        public int viewToModel(float n, float n2, Shape shape, Position.Bias[] array) {
            array[0] = Position.Bias.Forward;
            Rectangle rectangle = (Rectangle)shape;
            int n3 = (int)n;
            int n4 = (int)n2;
            if (n4 < rectangle.y) {
                return this.getStartOffset();
            }
            if (n4 > rectangle.y + rectangle.height) {
                return this.getEndOffset() - 1;
            }
            rectangle.height = WrappedPlainView.this.metrics.getHeight();
            int n5 = (rectangle.height > 0) ? ((n4 - rectangle.y) / rectangle.height) : (this.lineCount - 1);
            if (n5 >= this.lineCount) {
                return this.getEndOffset() - 1;
            }
            int startOffset = this.getStartOffset();
            int endOffset;
            if (this.lineCount == 1) {
                endOffset = this.getEndOffset();
            }
            else {
                int[] lineEnds = this.getLineEnds();
                endOffset = startOffset + lineEnds[n5];
                if (n5 > 0) {
                    startOffset += lineEnds[n5 - 1];
                }
            }
            if (n3 < rectangle.x) {
                return startOffset;
            }
            if (n3 > rectangle.x + rectangle.width) {
                return endOffset - 1;
            }
            Segment sharedSegment = SegmentCache.getSharedSegment();
            WrappedPlainView.this.loadText(sharedSegment, startOffset, endOffset);
            int tabbedTextOffset = Utilities.getTabbedTextOffset(sharedSegment, WrappedPlainView.this.metrics, rectangle.x, n3, WrappedPlainView.this, startOffset);
            SegmentCache.releaseSharedSegment(sharedSegment);
            return Math.min(startOffset + tabbedTextOffset, endOffset - 1);
        }
        
        @Override
        public void insertUpdate(DocumentEvent documentEvent, Shape shape, ViewFactory viewFactory) {
            this.update(documentEvent, shape);
        }
        
        @Override
        public void removeUpdate(DocumentEvent documentEvent, Shape shape, ViewFactory viewFactory) {
            this.update(documentEvent, shape);
        }
        
        private void update(DocumentEvent documentEvent, Shape shape) {
            int lineCount = this.lineCount;
            this.breakLines(documentEvent.getOffset());
            if (lineCount != this.lineCount) {
                WrappedPlainView.this.preferenceChanged(this, false, true);
                this.getContainer().repaint();
            }
            else if (shape != null) {
                Container container = this.getContainer();
                Rectangle rectangle = (Rectangle)shape;
                container.repaint(rectangle.x, rectangle.y, rectangle.width, rectangle.height);
            }
        }
        
        final int[] getLineEnds() {
            if (this.lineCache == null) {
                return null;
            }
            int[] array = (int[])this.lineCache.get();
            if (array == null) {
                return this.breakLines(this.getStartOffset());
            }
            return array;
        }
        
        final int[] breakLines(int n) {
            int[] referent = ((this.lineCache == null) ? null : ((int[])this.lineCache.get()));
            int startOffset = this.getStartOffset();
            int line = 0;
            if (referent != null) {
                line = this.findLine(n - startOffset);
                if (line > 0) {
                    --line;
                }
            }
            int i = (line == 0) ? startOffset : (startOffset + referent[line - 1]);
            int endOffset = this.getEndOffset();
            while (i < endOffset) {
                int calculateBreakPosition = WrappedPlainView.this.calculateBreakPosition(i, endOffset);
                i = ((calculateBreakPosition == i) ? (++calculateBreakPosition) : calculateBreakPosition);
                if (line == 0 && i >= endOffset) {
                    this.lineCache = null;
                    referent = null;
                    line = 1;
                    break;
                }
                if (referent == null || line >= referent.length) {
                    int[] array = new int[Math.max((int)Math.ceil((double)(line + 1) * ((double)(endOffset - startOffset) / (double)(i - startOffset))), line + 2)];
                    if (referent != null) {
                        System.arraycopy(referent, 0, array, 0, line);
                    }
                    referent = array;
                }
                referent[line++] = i - startOffset;
            }
            this.lineCount = line;
            if (this.lineCount > 1) {
                int n2 = this.lineCount + this.lineCount / 3;
                if (referent.length > n2) {
                    int[] array2 = new int[n2];
                    System.arraycopy(referent, 0, array2, 0, this.lineCount);
                    referent = array2;
                }
            }
            if (referent != null && referent != referent) {
                this.lineCache = new SoftReference(referent);
                return referent;
            }
            return referent;
        }
        
        private int findLine(int n) {
            int[] array = (int[])this.lineCache.get();
            if (n < array[0]) {
                return 0;
            }
            if (n > array[this.lineCount - 1]) {
                return this.lineCount;
            }
            return this.findLine(array, n, 0, this.lineCount - 1);
        }
        
        private int findLine(int[] array, int n, int n2, int n3) {
            if (n3 - n2 <= 1) {
                return n3;
            }
            int n4 = (n3 + n2) / 2;
            return (n < array[n4]) ? this.findLine(array, n, n2, n4) : this.findLine(array, n, n4, n3);
        }
    }
}
